@@ -1,5 +1,5 @@
 /*
-** pnscan.c - Manage the data connection
+** pnscan.c - Parallell Network Scanner
 **
 ** Copyright (c) 2002 Peter Eriksson <pen@lysator.liu.se>
 **
@@ -21,6 +21,7 @@
 #include <fcntl.h>
 #include <poll.h>
 #include <netdb.h>
+#include <locale.h>
 
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -37,15 +38,32 @@
 #include "bm.h"
 
 
+#define RESPONSE_MAX_SIZE    1024
+
+
 extern char version[];
 
-char *wstr = NULL;
-char *rstr = NULL;
+unsigned char *wstr = NULL;
+int   wlen = 0;
+unsigned char *rstr = NULL;
+int   rlen = 0;
 
+int debug = 0;
 int verbose = 0;
 
-int timeout = 1000;
+int stop = 0;
+
+int tworkers = 1;   /* Currently running worker threads */
+int mworkers = 1;   /* Peak started concurrent worker threads */
+int aworkers = 0;   /* Currently active probing worker threads */
+int nworkers = 0;   /* Max concurrent worker threads */
+
+int timeout = 1000; /* ms */
 int pr_sym  = 0;
+int line_f  = 0;
+int use_shutdown = 0;
+int maxlen  = 64;
+
 
 int first_port = 0;
 int last_port  = 0;
@@ -59,6 +77,9 @@ int cur_port;
 
 pthread_mutex_t print_lock;
 
+int ignore_case = 0;
+BM bmb;
+    
 
 void
 print_version(FILE *fp)
@@ -69,6 +90,135 @@ print_version(FILE *fp)
 }
 
 
+int
+get_char_code(unsigned char **cp,
+	      int base)
+{
+    int val = 0;
+    int len = 0;
+    
+    
+    while (len < (base == 16 ? 2 : 3) &&
+	   ((**cp >= '0' && **cp < '0'+(base > 10 ? 10 : base)) ||
+	    (base >= 10 && toupper(**cp) >= 'A' && toupper(**cp) < 'A'+base-10)))
+    {
+	val *= base;
+
+	if (**cp >= '0' && **cp < '0'+(base > 10 ? 10 : base))
+	    val += **cp - '0';
+	else if (base >= 10 &&
+		 toupper(**cp) >= 'A' && toupper(**cp) < 'A'+base-10)
+	    val += toupper(**cp) - 'A' + 10;
+
+	++*cp;
+	++len;
+    }
+
+    return val & 0xFF;
+}
+
+
+int
+dehex(unsigned char *str)
+{
+    unsigned char *wp, *rp;
+    int val;
+
+
+    rp = wp = str;
+
+    while (*rp)
+    {
+	while (*rp && isspace(* (unsigned char *) rp))
+	    ++rp;
+
+	if (*rp == '\0')
+	    break;
+	
+	if (!isxdigit(* (unsigned char *) rp))
+	    return -1;
+	
+	val = get_char_code(&rp, 16);
+	*wp++ = val;
+    }
+
+    *wp = '\0';
+    return wp - str;
+}
+
+
+    
+int
+deslash(unsigned char *str)
+{
+    unsigned char *wp, *rp;
+
+    
+    rp = wp = str;
+
+    while (*rp)
+    {
+	if (*rp != '\\')
+	    *wp++ = *rp++;
+	else
+	{
+	    switch (*++rp)
+	    {
+	      case 'n':
+		*wp++ = 10;
+		++rp;
+		break;
+
+	      case 'r':
+		*wp++ = 13;
+		++rp;
+		break;
+
+	      case 't':
+		*wp++ = 9;
+		++rp;
+		break;
+
+	      case 'b':
+		*wp++ = 8;
+		++rp;
+		break;
+
+	      case 'x':
+		++rp;
+		*wp++ = get_char_code(&rp, 16);
+		break;
+		
+	      case '0':
+		*wp++ = get_char_code(&rp, 8);
+		break;
+		
+	      case '1':
+	      case '2':
+	      case '3':
+	      case '4':
+	      case '5':
+	      case '6':
+	      case '7':
+	      case '8':
+	      case '9':
+		*wp++ = get_char_code(&rp, 10);
+		break;
+
+	      default:
+		*wp++ = *rp++;
+		break;
+	    }
+	}
+    }
+
+    *wp = '\0';
+
+    return wp-str;
+}
+
+
+
 void
 print_host(FILE *fp,
 	   struct in_addr in,
@@ -90,7 +240,7 @@ print_host(FILE *fp,
 
 int
 t_write(int fd,
-	char *buf,
+	unsigned char *buf,
 	int len)
 {
     int tw, wl, code;
@@ -129,7 +279,7 @@ t_write(int fd,
 
 int
 t_read(int fd,
-       char *buf,
+       unsigned char *buf,
        int size)
 {
     int len, code;
@@ -156,42 +306,49 @@ t_read(int fd,
 }
 
 int
-is_text(char *str)
+is_text(unsigned char *cp, int slen)
 {
-    unsigned char *cp = (unsigned char *) str;
-    
-    while (*cp && (isprint(*cp) || *cp == '\t' || *cp == '\n' || *cp == '\r'))
+    while (slen > 0 && (isprint(*cp) || *cp == '\0' || *cp == '\t' || *cp == '\n' || *cp == '\r'))
+    {
+	--slen;
 	++cp;
+    }
 
-    return *cp == 0;
+    return slen == 0;
 }
 
 
 int
-print_output(char *str)
+print_output(unsigned char *str, int slen)
 {
-    unsigned char *cp = (unsigned char *) str;
+    unsigned char *cp = str;
+    int len;
+    
 
+    len = 0;
     
     if (str == NULL)
     {
-	printf("TXT :");
-	return -1;
+	printf("NULL");
+	return len;
     }
 
-    if (cp[0] == IAC && cp[1] >= xEOF)
+
+    if (slen >= 2 && cp[0] == IAC && cp[1] >= xEOF)
     {
 	printf("TEL : ");
 
-	while (*cp)
+	while (len < slen && len < maxlen)
 	{
 	    if (*cp == IAC)
 	    {
+		++len;
+		
 		printf("<IAC>");
 		switch (*++cp)
 		{
 		  case 0:
-		    return 0;
+		    return len;
 
 		  case DONT:
 		    printf("<DONT>");
@@ -273,41 +430,72 @@ print_output(char *str)
 		    printf("<0x%02X>", *cp);
 		}
 	    }
+	    
 	    else if (isprint(*cp))
 		putchar(*cp);
+	    
 	    else
+	    {
 		switch (*cp)
 		{
 		  case '\n':
+		    if (line_f)
+			return len;
+
+		    printf("\\n");
+		    break;
+		    
 		  case '\r':
-		    return 0;
+		    if (line_f)
+			return len;
+		    
+		    printf("\\r");
+		    break;
 		    
 		  case '\t':
 		    printf("\\t");
 		    break;
+
+		  case '\0':
+		    printf("\\0");
+		    break;
 		    
 		  default:
 		    printf("\\x%02X", *cp);
 		}
-	    
+	    }
+
+	    ++len;
 	    ++cp;
 	}
     }
 
-    else if (is_text(str))
+    else if (is_text(str, slen))
     {
 	printf("TXT : ");
     
-	while (*str)
+	while (len < slen && len < maxlen)
 	{
 	    if (isprint(* (unsigned char *) str))
 		putchar(*str);
 	    else
 		switch (*str)
 		{
+		  case '\0':
+		    printf("\\0");
+		    break;
+		    
 		  case '\n':
+		    if (line_f)
+			return len;
+		    printf("\\n");
+		    break;
+		    
 		  case '\r':
-		    return 0;
+		    if (line_f)
+			return len;
+		    printf("\\r");
+		    break;
 		    
 		  case '\t':
 		    printf("\\t");
@@ -316,24 +504,24 @@ print_output(char *str)
 		  default:
 		    printf("\\x%02x", * (unsigned char *) str);
 		}
-	    
+
+	    ++len;
 	    ++str;
 	}
     }
     
     else
     {
-	int len = 0;
-	
 	printf("HEX :");
-	while (*str && ++len < 20)
+	while (len < slen && len < maxlen)
 	{
 	    printf(" %02x", * (unsigned char *) str);
+	    ++len;
 	    ++str;
 	}
     }
 
-    return 0;
+    return len;
 }
 
 
@@ -344,7 +532,7 @@ probe(unsigned long addr,
 {
     int fd, code, len;
     struct sockaddr_in sin;
-    char buf[512];
+    unsigned char buf[RESPONSE_MAX_SIZE];
     struct pollfd pfd;
     
 
@@ -416,7 +604,7 @@ probe(unsigned long addr,
 
     if (wstr)
     {
-	code = t_write(fd, wstr, strlen(wstr));
+	code = t_write(fd, wstr, wlen);
 	if (code < 0)
 	{
 	    if (verbose)
@@ -432,30 +620,14 @@ probe(unsigned long addr,
 	    close(fd);
 	    return 0;
 	}
-	
-	code = t_write(fd, "\r\n\r\n", 4);
-	if (code < 0)
-	{
-	    if (verbose)
-	    {
-		pthread_mutex_lock(&print_lock);
-		
-		print_host(stderr, sin.sin_addr, port);
-		fprintf(stderr, " : ERR : write() failed: %s\n", strerror(errno));
-		
-		pthread_mutex_unlock(&print_lock);
-	    }
-	    
-	    close(fd);
-	    return 0;
-	}
     }
 
-    shutdown(fd, 1);
+    if (use_shutdown)
+	shutdown(fd, 1);
 
     while ((len = t_read(fd, buf, sizeof(buf)-1)) < 0 && errno == EINTR)
 	;
-    
+
     if (len < 0)
     {
 	if (verbose)
@@ -474,34 +646,25 @@ probe(unsigned long addr,
 
     buf[len] = '\0';
 	    
-    if (len == 0)
-    {
-	pthread_mutex_lock(&print_lock);
-	
-	print_host(stdout, sin.sin_addr, port);
-	printf(" : NULL\n");
-	pthread_mutex_unlock(&print_lock);
-	
-	close(fd);
-	return 1;
-    }
-
     if (rstr)
     {
 	int pos;
 
-	pos = bm_search(buf, len);
+	pos = bm_search(&bmb, buf, len, NULL);
 
 	if (pos >= 0)
 	{
-	    while (pos > 0 && !(buf[pos-1] == '\n' || buf[pos-1] == '\r'))
-		--pos;
+	    if (line_f)
+		while (pos > 0 &&
+		       !(buf[pos-1] == '\n' || buf[pos-1] == '\r'))
+		    --pos;
 	    
 	    pthread_mutex_lock(&print_lock);
 	    
 	    print_host(stdout, sin.sin_addr, port);
 	    printf(" : ");
-	    print_output(buf+pos);
+
+	    print_output(buf+pos, len-pos);
 	    putchar('\n');
 	    
 	    pthread_mutex_unlock(&print_lock);
@@ -513,7 +676,7 @@ probe(unsigned long addr,
 	
 	print_host(stdout, sin.sin_addr, port);
 	printf(" : ");
-	print_output(buf);
+	print_output(buf, len);
 	putchar('\n');
 	
 	pthread_mutex_unlock(&print_lock);
@@ -529,33 +692,58 @@ r_worker(void *arg)
 {
     unsigned long addr;
     int port;
+    pthread_t tid;
     
     
-  Again:
     pthread_mutex_lock(&cur_lock);
-
-    if (cur_ip <= last_ip)
+	
+    while (!stop)
     {
-	port = cur_port;
-	addr = cur_ip++;
-    }
-    else
-    {
-    	if (cur_port >= last_port)
+	if (cur_ip <= last_ip)
 	{
-	    pthread_mutex_unlock(&cur_lock);
-	    return NULL;
+	    port = cur_port;
+	    addr = cur_ip++;
+	}
+	else
+	{
+	    if (cur_port >= last_port)
+	    {
+		stop = 1;
+		break;
+	    }
+	    
+	    port = ++cur_port;
+	    addr = cur_ip = first_ip;
 	}
 
-	port = ++cur_port;
-	addr = cur_ip = first_ip;
+	if (aworkers >= tworkers-1 && tworkers < nworkers)
+	{
+	    ++tworkers;
+
+	    if (pthread_create(&tid, NULL, r_worker, NULL) != 0)
+	    {
+		--tworkers;
+		nworkers = tworkers;
+	    }
+	    
+	    if (tworkers > mworkers)
+		mworkers = tworkers;
+	}
+
+	++aworkers;
+	pthread_mutex_unlock(&cur_lock);
+
+	probe(addr, port);
+	
+	pthread_mutex_lock(&cur_lock);
+	--aworkers;
     }
+
+    --tworkers;
     
     pthread_mutex_unlock(&cur_lock);
-
-    probe(addr, port);
-    
-    goto Again;
+    fflush(stdout); 
+    return NULL;
 }
 
 
@@ -608,54 +796,89 @@ f_worker(void *arg)
     unsigned long addr;
     int port, code;
     char buf[1024];
-    char host[1024];
-    char serv[256];
+    char *host;
+    char *serv;
+    char *tokp;
+    pthread_t tid;
     
     
-  Again:
     pthread_mutex_lock(&cur_lock);
-
-    host[0] = serv[0] = '\0';
     
-    if (feof(stdin) || fgets(buf, sizeof(buf), stdin) == NULL)
+    while (!stop)
     {
+	if (fgets(buf, sizeof(buf), stdin) == NULL)
+	{
+	    if (debug)
+		fprintf(stderr, "*** GOT EOF ***\n");
+	    
+	    stop = 1;
+	    break;
+	}
+
+	host = strtok_r(buf, " \t\n\r", &tokp);
+	serv = strtok_r(NULL, " \t\n\r", &tokp);
+	
+	if (host == NULL || host[0] == '#')
+	    continue;
+	
+	if (get_host(host, &addr) != 1)
+	{
+	    if (verbose)
+		fprintf(stderr, "%s: invalid host\n", host);
+	    continue;
+	}
+	
+	if (serv == NULL)
+	{
+	    if (first_port == 0)
+	    {
+		if (verbose)
+		    fprintf(stderr, "%s: missing service specification\n",
+			    host);
+		continue;
+	    }
+
+	    port = first_port;
+	}
+	else
+	{
+	    code = get_service(serv, &port);
+	    if (code != 1)
+	    {
+		if (verbose)
+		    fprintf(stderr, "%s: invalid service (code=%d)\n", serv, code);
+		continue;
+	    }
+	}
+	
+	if (aworkers >= tworkers-1 && tworkers < nworkers)
+	{
+	    ++tworkers;
+	    
+	    if (pthread_create(&tid, NULL, f_worker, NULL) != 0)
+	    {
+		--tworkers;
+		nworkers = tworkers;
+	    }
+
+	    if (tworkers > mworkers)
+		mworkers = tworkers;
+	}
+
+	++aworkers;
 	pthread_mutex_unlock(&cur_lock);
-	return NULL;
+	
+	probe(addr, port);
+	
+	pthread_mutex_lock(&cur_lock);
+	--aworkers;
     }
 
+    --tworkers;
+    
     pthread_mutex_unlock(&cur_lock);
-    
-    host[0] = serv[0] = '\0';
-    code = sscanf(buf, "%1024s %255s", host, serv);
-
-    if (code >= 1 && host[0] == '#')
-	goto Again;
-    
-    if (code < 2)
-    {
-	if (code > 0 && verbose)
-	    fprintf(stderr, "%s%s%s: invalid host & service\n",
-		    host, serv[0] ? " " : "", serv);
-	goto Again;
-    }
-
-    if (get_host(host, &addr) != 1)
-    {
-	if (verbose)
-	    fprintf(stderr, "%s: invalid host\n", host);
-	goto Again;
-    }
-
-    code = get_service(serv, &port);
-    if (code != 1)
-    {
-	if (verbose)
-	    fprintf(stderr, "%s: invalid service (code=%d)\n", serv, code);
-	goto Again;
-    }
-
-    probe(addr, port);
-    goto Again;
+    fflush(stdout); 
+    return NULL;
 }
 
 
@@ -665,15 +888,30 @@ char *argv0 = "pnscan";
 void
 usage(FILE *out)
 {
-    fprintf(out, "Usage: %s [<options>] <CIDR | host-range> <port-range>\n", argv0);
-    fprintf(out, "Options:\n");
-    fprintf(out, "\t-h                Display this information.\n");
-    fprintf(out, "\t-s                Lookup and print hostnames.\n");
-    fprintf(out, "\t-v                Be verbose.\n");
-    fprintf(out, "\t-w<string>        Request string to send.\n");
-    fprintf(out, "\t-r<string>        Response string to look for.\n");
-    fprintf(out, "\t-t<msecs>         Connect/Write/Read timeout.\n");
-    fprintf(out, "\t-n<workers>       Concurrent worker threads.\n");
+    fprintf(out, "Usage: %s [<options>] [{<CIDR>|<host-range> <port-range>} | <service>]\n", argv0);
+
+    fputs("\n\
+This program implements a multithreaded TCP port scanner.\n\
+More information may be found at:\n\
+\thttp://www.lysator.liu.se/~pen/pnscan\n\
+\n\
+Command line options:\n", out);
+    
+    fprintf(out, "\t-h             Display this information.\n");
+    fprintf(out, "\t-V             Print version.\n");
+    fprintf(out, "\t-v             Be verbose.\n");
+    fprintf(out, "\t-d             Print debugging info.\n");
+    fprintf(out, "\t-s             Lookup and print hostnames.\n");
+    fprintf(out, "\t-i             Ignore case when scanning responses.\n");
+    fprintf(out, "\t-S             Enable shutdown mode.\n");
+    fprintf(out, "\t-l             Line oriented output.\n");
+    fprintf(out, "\t-w<string>     Request string to send.\n");
+    fprintf(out, "\t-W<hex list>   Hex coded request string to send.\n");
+    fprintf(out, "\t-r<string>     Response string to look for.\n");
+    fprintf(out, "\t-R<hex list>   Hex coded response string to look for.\n");
+    fprintf(out, "\t-L<length>     Max bytes to print.\n");
+    fprintf(out, "\t-t<msecs>      Connect/Write/Read timeout.\n");
+    fprintf(out, "\t-n<workers>    Concurrent worker threads limit.\n");
 }
     
 
@@ -702,13 +940,13 @@ get_ip_range(char *str,
 	     unsigned long *first_ip,
 	     unsigned long *last_ip)
 {
-    char first[256], last[256];
+    char first[1024], last[1024];
     int len;
     unsigned long ip;
     unsigned long mask = 0;
     
 
-    if (sscanf(str, "%255[^/ ] / %u", first, &len) == 2)
+    if (sscanf(str, "%1023[^/ ] / %u", first, &len) == 2)
     {
 	/* CIDR */
 
@@ -727,7 +965,7 @@ get_ip_range(char *str,
 	return 2;
     }
 
-    switch (sscanf(str, "%255[^: ] : %255s", first, last))
+    switch (sscanf(str, "%1023[^: ] : %1023s", first, last))
     {
       case 1:
 	if (get_host(first, first_ip) != 1)
@@ -788,17 +1026,27 @@ get_port_range(char *str,
 }
 
 
+
+void
+e_fun(void)
+{
+    printf("mworkers = %d, tworkers = %d, aworkers = %d, nworkers = %d\n",
+	   mworkers, tworkers, aworkers, nworkers);
+}
+
+
 int
 main(int argc,
      char *argv[])
 {
-    int i, nworkers, len;
-    pthread_t tid;
+    int i, j;
     struct rlimit rlb;
+    char *arg;
+    
     
-
     argv0 = argv[0];
 
+    setlocale(LC_CTYPE, "");
 
     first_port = 0;
     last_port = 0;
@@ -815,109 +1063,191 @@ main(int argc,
 	nworkers = 1024;
 
     pthread_mutex_init(&cur_lock, NULL);
+    pthread_mutex_init(&print_lock, NULL);
 
     for (i = 1; i < argc && argv[i][0] == '-'; i++)
-	switch (argv[i][1])
-	{
-	  case '-':
-	    ++i;
-	    goto EndOptions;
-	    
-	  case 'h':
-	    usage(stdout);
-	    exit(0);
-    
-	  case 'w':
-	    wstr = argv[i]+2;
-	    break;
-
-	  case 's':
-	    ++pr_sym;
-	    break;
-	    
-	  case 'r':
-	    rstr = argv[i]+2;
-	    if (bm_setup(rstr, strlen(rstr)) < 0)
+	for (j = 1; j > 0 && argv[i][j]; ++j)
+	    switch (argv[i][j])
 	    {
-		fprintf(stderr, "%s: Failed search string setup: %s\n",
-			argv[0], argv[i]+2);
+	      case '-':
+		++i;
+		goto EndOptions;
+		
+	      case 'V':
+		print_version(stdout);
+		break;
+
+	      case 'd':
+		++debug;
+		break;
+
+	      case 'i':
+		ignore_case = 1;
+		break;
+		
+	      case 'v':
+		++verbose;
+		break;
+		
+	      case 'h':
+		usage(stdout);
+		exit(0);
+		
+	      case 'l':
+		++line_f;
+		break;
+		
+	      case 's':
+		++pr_sym;
+		break;
+		
+	      case 'S':
+		++use_shutdown;
+		break;
+		
+	      case 'w':
+		if (argv[i][2])
+		    wstr = (unsigned char *) strdup(argv[i]+2);
+		else
+		    wstr = (unsigned char *) strdup(argv[++i]);
+		
+		wlen = deslash(wstr);
+		j = -2;
+		break;
+
+	      case 'W':
+		if (argv[i][2])
+		    wstr = (unsigned char *) strdup(argv[i]+2);
+		else
+		    wstr = (unsigned char *) strdup(argv[++i]);
+		wlen = dehex(wstr);
+		j = -2;
+		break;
+	    
+	      case 'R':
+		if (argv[i][2])
+		    rstr = (unsigned char *) strdup(argv[i]+2);
+		else
+		    rstr = (unsigned char *) strdup(argv[++i]);
+		rlen = dehex(rstr);
+		j = -2;
+		break;
+		
+	      case 'r':
+		if (argv[i][2])
+		    rstr = (unsigned char *) strdup(argv[i]+2);
+		else
+		    rstr = (unsigned char *) strdup(argv[++i]);
+		rlen = deslash(rstr);
+		j = -2;
+		break;
+		
+	      case 'L':
+		if (argv[i][2])
+		    arg = argv[i]+2;
+		else
+		    arg = argv[++i];
+		
+		if (!arg || sscanf(arg, "%u", &maxlen) != 1)
+		{
+		    fprintf(stderr, "%s: Invalid length specification: %s\n",
+			    argv[0], arg ? arg : "<null>");
+		    exit(1);
+		}
+		j = -2;
+		break;
+		
+	      case 't':
+		if (argv[i][2])
+		    arg = argv[i]+2;
+		else
+		    arg = argv[++i];
+		
+		if (!arg || sscanf(arg, "%u", &timeout) != 1)
+		{
+		    fprintf(stderr,
+			    "%s: Invalid timeout specification: %s\n",
+			    argv[0], arg ? arg : "<null>");
+		    exit(1);
+		}
+		j = -2;
+		break;
+		
+	      case 'n':
+		if (argv[i][2])
+		    arg = argv[i]+2;
+		else
+		    arg = argv[++i];
+		
+		if (!arg || sscanf(arg, "%u", &nworkers) != 1)
+		{
+		    fprintf(stderr,
+			    "%s: Invalid workers specification: %s\n",
+			    argv[0], arg ? arg : "<null>");
+		    exit(1);
+		}
+		j = -2;
+		break;
+		
+	      default:
+		fprintf(stderr, "%s: unknown command line switch: -%c\n",
+			argv[0], argv[i][j]);
 		exit(1);
 	    }
-	    break;
-
-	  case 't':
-	    timeout = atoi(argv[i]+2);
-	    break;
-
-	  case 'n':
-	    nworkers = atoi(argv[i]+2);
-	    break;
-
-	  case 'V':
-	    print_version(stdout);
-	    break;
-	    
-	  case 'v':
-	    ++verbose;
-	    break;
-	    
-	  default:
-	    fprintf(stderr, "%s: unknown command line switch: %s\n",
-		    argv[0], argv[i]);
-	    exit(1);
-	}
 
   EndOptions:
-    
-    if (i == argc)
+
+    if (rstr)
     {
-	for (i = 0; i < nworkers && !feof(stdin); ++i)
-	    pthread_create(&tid, NULL, f_worker, NULL);
-	
-	pthread_exit(NULL);
+	if (bm_init(&bmb, rstr, rlen, ignore_case) < 0)
+	{
+	    fprintf(stderr, "%s: Failed search string setup: %s\n",
+		    argv[0], rstr);
+	    exit(1);
+	}
     }
 
-    else
+    if (debug)
+	atexit(e_fun);
+	
+    if (i == argc || i+1 == argc)
     {
-	
-	if (i + 2 != argc)
-	{
-	    fprintf(stderr,
-		    "%s: Missing or extra argument(s). Use '-h' for help.\n",
-		    argv[0]);
-	    exit(1);
-	}
-	
-	if (get_ip_range(argv[i], &first_ip, &last_ip) < 1)
-	{
-	    fprintf(stderr, "%s: Invalid IP address range: %s\n",
-		    argv[0], argv[i]);
-	    exit(1);
-	}
-	
-	if (get_port_range(argv[i+1], &first_port, &last_port) < 1)
-	{
-	    fprintf(stderr, "%s: Invalid Port range: %s\n",
-		    argv[0], argv[i+1]);
-	    exit(1);
-	}
-	
-	cur_ip = first_ip;
-	cur_port = first_port;
-	
-	len = (last_ip - first_ip + 1)*(last_port - first_port + 1);
-	
-	if (len < nworkers)
-	    nworkers = len;
-	
-	if (verbose > 1)
-	    fprintf(stderr, "Using %d worker threads\n", nworkers);
-	
-	for (i = 0; i < nworkers; ++i)
-	    pthread_create(&tid, NULL, r_worker, NULL);
-	
+	if (i + 1 == argc)
+	    get_service(argv[i], &first_port);
+			  
+	f_worker(NULL);
 	pthread_exit(NULL);
-    }
 	
+	return 1; /* Not reached */
+    }
+
+    if (i + 2 != argc)
+    {
+	fprintf(stderr,
+		"%s: Missing or extra argument(s). Use '-h' for help.\n",
+		argv[0]);
+	exit(1);
+    }
+    
+    if (get_ip_range(argv[i], &first_ip, &last_ip) < 1)
+    {
+	fprintf(stderr, "%s: Invalid IP address range: %s\n",
+		argv[0], argv[i]);
+	exit(1);
+    }
+    
+    if (get_port_range(argv[i+1], &first_port, &last_port) < 1)
+    {
+	fprintf(stderr, "%s: Invalid Port range: %s\n",
+		argv[0], argv[i+1]);
+	exit(1);
+    }
+    
+    cur_ip = first_ip;
+    cur_port = first_port;
+
+    r_worker(NULL);
+    pthread_exit(NULL);
+    
     return 1; /* Not reached */
 }
